/*
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <pthread.h>

#include <snuff.h>
#include <snuff_list.h>
#include <snuff_util.h>
#include <snuff_mta.h>
#include <snuff_db.h>

extern u_int32_t max_aliases;

static pthread_mutex_t snuff_lock = PTHREAD_MUTEX_INITIALIZER;
static hash_t alias_list;
static hash_t user_list;

static int __delete_alias(uid_t uid, const char *username);

/*	  
** Hash by Karl Nelson <kenelson@ece.ucdavis.edu>
**
** See http://mail.gnome.org/archives/gtk-devel-list/2000-February/msg00057.html
** for details. 
*/  
		
static inline u_int32_t alias_hash(const char *str) {
	u_int32_t hash = 0;
	
	while (*str != '\0')
		hash = (hash << 5) - hash + *str++;

	hash &= (1 << USER_HASH_ORDER) - 1;
	return (hash);
}

static inline u_int32_t user_hash(uid_t uid) {
	return ((u_int32_t) uid & (((1 << USER_HASH_ORDER)) - 1));
}

static int alias_compare(void *l, void *r) {
	return (strcasecmp(l, r));
}

static void alias_remove(void *alias_str) {
	if (del_mta_alias((char *) alias_str) != 0) {
		debug("Error removing MTA alias: %s\n",
			strerror(errno));
	}

	free(alias_str);
}

static int user_compare(void *l, void *r) {
	struct user_record *left = l;
	struct user_record *right = r;

	return (left->uid - right->uid);
}

/*
** Must be called with snuff_lock held.
*/

static void user_remove(void *rec) {
	struct user_record *user_record = rec;
	dlist_t *cur;

	cur = user_record->alias_list;
	while (cur != NULL) {
		struct alias_block *alias = cur->data;
		dlist_t *next = cur->next;
		int ret;

		/* Also remove the username from the user hash */
		ret = hash_remove(&alias_list, alias->username,
				alias_hash(alias->username));

		if (ret != 0)
			debug("Error removing username \"%s\" from alias hash", alias->username);

		free(cur->data);
		free(cur);

		cur = next;
	}

	free(rec);
}

static char *generate_new_alias(size_t len) {
	char *user;
	size_t i;
	static const char valid_chars[] = "abcdefghijklmnopqrstuuvwxyz";

	user = xmalloc(len);

	for (i = 0 ; i < len - 1 ; i++)
		user[i] = valid_chars[rand() % (sizeof(valid_chars) - 1)];

	user[i] = '\0';
	return (user);
}

static struct user_record *get_user_record(uid_t uid) {
	u_int32_t uhash;
	dlist_t *node;
	struct user_record temp;

	uhash = user_hash(uid);
	temp.uid = uid;

	node = hash_find(&user_list, &temp, uhash);
	if (node == NULL)
		return (NULL);

	return (node->data);
}

static int add_alias(uid_t uid, const char *username, time_t expire) {
	struct user_record *user_record;
	struct alias_block *alias;

	user_record = get_user_record(uid);
	if (user_record == NULL) {
		/*
		** There are no aliases for this user so far.
		*/
		user_record = xmalloc(sizeof(*user_record));
		user_record->uid = uid;
		user_record->alias_list = NULL;
		user_record->num_aliases = 0;

		hash_add(&user_list, user_record, user_hash(uid));
	}

	if (max_aliases > 0 && user_record->num_aliases >= max_aliases)
		return (-1);

	alias = xmalloc(sizeof(*alias));
	xstrncpy(alias->username, username, sizeof(alias->username));
	alias->expire = expire;

	user_record->alias_list = dlist_add_head(user_record->alias_list, alias);
	user_record->num_aliases++;

	return (0);
}

int init_state(void) {
	hash_init(&alias_list, USER_HASH_ORDER, alias_compare, alias_remove);
	hash_init(&user_list, USER_HASH_ORDER, user_compare, user_remove);

	/*
	** read in any saved state
	*/

	return (0);
}

int new_alias(uid_t uid, struct alias_block *alias, size_t len) {
	char *username;
	size_t i;
	u_int32_t hash = 0;
	char *login;
	int ret = 0;

	if (len > sizeof(alias->username))
		return (RESULT_TOOLONG);

	/*
	** Has to live for at least a minute.
	*/

	if (alias->expire < time(NULL) + 60)
		return (RESULT_EXPIRESOON);

	pthread_mutex_lock(&snuff_lock);

	/*
	** Try 10 times to generate a good username -- this should
	** succeed on the first try very close to 100% of the time.
	*/

	for (i = 0 ; i < 10 ; i++) {
		username = generate_new_alias(len);
		if (username == NULL)
			continue;

		hash = alias_hash(username);

		if (in_hash(&alias_list, username, hash)) {
			free(username);
			username = NULL;
			continue;
		}
	}

	if (username == NULL) {
		ret = RESULT_FAIL;
		goto out;
	}

	hash_add(&alias_list, username, hash);
	if (add_alias(uid, username, alias->expire) == -1) {
		ret = RESULT_TOOMANY;
		goto out;
	}

	login = get_username(uid);
	ret = add_mta_alias(username, login);
	free(login);

	if (ret != 0) {
		__delete_alias(uid, username);
		ret = RESULT_FAIL;
		goto out;
	}

	xstrncpy(alias->username, username, sizeof(alias->username));

out:
	pthread_mutex_unlock(&snuff_lock);
	return (ret);
}

/*
** snuff_lock must be held when this function is called.
*/

static int __delete_alias(uid_t uid, const char *username) {
	struct user_record *user_record;
	dlist_t *cur;

	user_record = get_user_record(uid);
	if (user_record == NULL)
		return (RESULT_NOTEXIST);

	cur = user_record->alias_list;
	while (cur != NULL) {
		dlist_t *next = cur->next;
		struct alias_block *alias = cur->data;

		if (!strcasecmp(alias->username, username)) {
			int ret;

			ret = hash_remove(&alias_list, alias->username,
					alias_hash(alias->username));

			if (ret != 0)
				debug("Error removing from alias hash.");

			free(cur->data);

			user_record->alias_list =
				dlist_remove(user_record->alias_list, cur);
			user_record->num_aliases--;

			/*
			** If we've removed a user's last alias, delete the user
			** from the user list.
			*/

			if (user_record->alias_list == NULL) {
				hash_remove(&user_list,
						user_record, user_hash(user_record->uid));
			}

			return (0);
		}

		cur = next;
	}

	return (RESULT_NOTEXIST);
}

int delete_alias(uid_t uid, const char *username) {
	int ret;

	pthread_mutex_lock(&snuff_lock);
	ret = __delete_alias(uid, username);
	pthread_mutex_unlock(&snuff_lock);

	return (ret);
}

int delete_all_aliases(uid_t uid) {
	int ret;
	struct user_record temp;
	u_int32_t uhash = user_hash(uid);

	temp.uid = uid;

	pthread_mutex_lock(&snuff_lock);
	ret = hash_remove(&user_list, &temp, uhash);
	pthread_mutex_unlock(&snuff_lock);

	if (ret != 0)
		return (RESULT_NOALIASES);

	return (0);
}

int list_aliases(int sock, uid_t uid) {
	struct user_record *user_record;
	dlist_t *cur;
	int ret = 0;

	pthread_mutex_lock(&snuff_lock);

	user_record = get_user_record(uid);
	if (user_record == NULL) {
		ret = RESULT_NOTEXIST;
		goto out;
	}

	for (cur = user_record->alias_list ; cur != NULL ; cur = cur->next) {
		struct alias_block *alias = cur->data;
		int ret;

		ret = sock_write(sock, alias, sizeof(*alias));
		if (ret != sizeof(*alias)) {
			debug("Error: write: %s", strerror(errno));
			ret = RESULT_FAIL;
			goto out;
		}
	}

out:
	pthread_mutex_unlock(&snuff_lock);
	return (ret);
}

int save_state(const char *output_file) {
	FILE *fp;
	size_t i;
	u_int32_t order;

	fp = fopen(output_file, "w");
	if (fp == NULL) {
		debug("Error: open: %s", strerror(errno));
		return (-1);
	}

	if (fchmod(fileno(fp), 0600) != 0)
		debug("Error: chmod: %s", strerror(errno));

	pthread_mutex_lock(&snuff_lock);

	order = 1 << user_list.order;

	for (i = 0 ; i < order ; i++) {
		dlist_t *cur;

		for (cur = user_list.map[i] ; cur != NULL ; cur = cur->next) {
			dlist_t *cur_alias;
			struct user_record *user_record = cur->data;

			cur_alias = user_record->alias_list;

			for (; cur_alias != NULL ; cur_alias = cur_alias->next) {
				struct alias_block *alias = cur_alias->data;
				struct saved_record sr;
				int ret;

				sr.uid = user_record->uid;
				sr.expire = alias->expire;
				xstrncpy(sr.username, alias->username, sizeof(sr.username));

				ret = fwrite(&sr, sizeof(sr), 1, fp);
				if (ret != 1)
					debug("Error saving state: %s", strerror(errno));
			}
		}
	}

	pthread_mutex_unlock(&snuff_lock);
	fclose(fp);
	return (0);
}

/*
** This must be called before the server starts accepting connections.
*/

int read_state(const char *filename) {
	FILE *fp;
	struct saved_record sr;

	fp = fopen(filename, "r");
	if (fp == NULL) {
		debug("Error: open: %s", strerror(errno));
		return (-1);
	}

	while (1) {
		int ret = fread(&sr, sizeof(sr), 1, fp);
		char *username;

		if (ret != 1) {
			if (!feof(fp))
				debug("Error reading saved state: %s", strerror(errno));
			break;
		}

		if (check_mta_alias(sr.username) != 0) {
			/*
			** somehow the alias is gone, try to add it back.
			*/
			char *login;

			login = get_username(sr.uid);
			ret = add_mta_alias(sr.username, login);
			free(login);

			if (ret != 0) {
				debug("Error restoring alias \"%s\" for uid %d",
					sr.username, sr.uid);
				continue;
			}
		}

		username = xstrdup(sr.username);

		if (add_alias(sr.uid, sr.username, sr.expire) != -1)
			hash_add(&alias_list, username, alias_hash(username));
		else
			free(username);
	}

	fclose(fp);
	return (0);
}

void expire_aliases(void) {
	size_t i;
	u_int32_t order;

	pthread_mutex_lock(&snuff_lock);

	order = 1 << user_list.order;

	for (i = 0 ; i < order ; i++) {
		dlist_t *cur;

		cur = user_list.map[i];
		while (cur != NULL) {
			dlist_t *next = cur->next;
			dlist_t *cur_alias;
			struct user_record *user_record = cur->data;

			cur_alias = user_record->alias_list;
			while (cur_alias != NULL) {
				dlist_t *next_alias = cur_alias->next;
				struct alias_block *alias = cur_alias->data;

				if (alias->expire <= time(NULL)) {
					int ret;
					ret = __delete_alias(user_record->uid, alias->username);

					if (ret != 0)
						debug("Error: delete_alias: %d", ret);
				}

				cur_alias = next_alias;
			}

			cur = next;
		}
	}

	pthread_mutex_unlock(&snuff_lock);
}
